/*
	Copyright 2007 Coraid, Inc.

	Author:
	Sam Hopkins
	sah@coraid.com

	We define _FILE_OFFSET_BITS to get 64 bit pread/pwrite support.
	We define _GNU_SOURCE to obtain O_LARGEFILE
*/

#define _FILE_OFFSET_BITS 64
#define _GNU_SOURCE
#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>
#include "p9p.h"
#include "ddt.h"

// OSX BSD doesn't support O_LARGEFILE as it's all 64 bit 
#ifndef O_LARGEFILE
#define O_LARGEFILE 0
#endif

enum {
	/* 4GB io default */
	Dbs= 256*1024,
	Dcount= 16*1024,
	Dssecs= 10,

	MAXPT= 64
};

pthread_t threads[MAXPT];
pthread_mutex_t iomutex = PTHREAD_MUTEX_INITIALIZER;
int nthreads = 1;
unsigned long long nextio;
char file[4096];
char *block;
static int iofd;
unsigned long long bs = Dbs;
unsigned long long count = Dcount;
unsigned long long total;	// can't init here, "integer overflow expression" (*TI)
int fflag;
int Uflag;
int rflag;
int wflag;
int cflag;
int dflag;
int ssecs = Dssecs;
void *rstats, *wstats;

void *
iothread(void *vp)
{
	unsigned long long i;
	long long n;
	ssize_t (*rw)(int, void *, size_t, off_t) = vp;

	for (;;) {
		pthread_mutex_lock(&iomutex);
		i = nextio++;
		pthread_mutex_unlock(&iomutex);
		if (i >= count)
			break;
		n = rw(iofd, block, bs, bs*i);
		if (n != bs)
			sysfatal("io error %lld/%lld at offset %lld\n", n, bs, bs*i);
	}
	return 0;
}

void
threadio(ssize_t (*rw)(int, void *, size_t, off_t))
{
	int n, i;

	nextio = 0;
	if (nthreads <= 1) {
		iothread(rw);
		return;
	}
	for (i=0; i<nthreads; i++) {
		n = pthread_create(&threads[i], 0, iothread, (void *) rw);
		if (n)
			fprintf(stderr, "error creating thread %d\n", i);
		else
			dprint("created thread %d:%lu\n", i, (unsigned long) threads[i]);
	}
	for (i=0; i<nthreads; i++) {
		dprint("joining thread %d:%lu\n", i, (unsigned long) threads[i]);
		pthread_join(threads[i], 0);
	}
}

void
ddr(void)
{
	fprintf(stderr, "Reading from %s ... ", file);
	rstats = stats(rstats);
	threadio(&pread);
	stats(rstats);
	fprintf(stderr, "done.\n");
}

void
ddw(void)
{
	fprintf(stderr, "Writing to %s ... ", file);
	wstats = stats(wstats);
	threadio((ssize_t (*)(int, void *, size_t, off_t)) &pwrite);  // shut it!
	fprintf(stderr, "syncing ... ");
	sync();
	stats(wstats);
	fprintf(stderr, "done.\n");
}

void
ddt(void)
{
	char buf[32];

	snprintf(buf, sizeof buf, "%lldMiB", (bs*count) >> 20LL);
	printf("%-8s  %7s  %4s\n", buf, "KiB/s", "CPU%");
	if (wflag) {
		printf("%-8s  ", "Write");
		cookedstats(wstats, bs, count);
	}
	if (rflag) {
		printf("%-8s  ", "Read");
		cookedstats(rstats, bs, count);
	}
	if (dflag) {
		printf("\n");
		printf("WRaw ");
		rawstats(wstats);
		printf("RRaw ");
		rawstats(rstats);
	}
}

void
usage(void)
{
	fprintf(stderr, "usage: ddt [-?] [-t total] dir\n"); 
	exit(1);
}

void *
balloc(unsigned long bs)
{
	char *p;
	size_t ps;
	int i;

	ps = getpagesize();
	p = malloc(bs+ps);
	p += ps - 1;
	p -= (size_t) p % ps;

	for (i=0; i<bs; i++)
		p[i] = i;
	return p;
}

/* formatted str to ll */
unsigned long long
fstrtoll(char *str)
{
	unsigned long long n;
	char *p;

	n = strtoll(str, &p, 0);
	if (n < 0 || p == str)
		goto error;
	switch (*p) {
	case 'g':
	case 'G':
		n <<= 10;
	case 'm':
	case 'M':
		n <<= 10;
	case 'K':
	case 'k':
		n <<= 10;
		p++;
	}
	if (*p != '\0')
error:		sysfatal("invalid number format %s\n", str);
	return n;
}

int
main(int argc, char *argv[])
{
	char *dir;
	int n;
	struct stat s;

	ARGBEGIN {
	case 'U':
		Uflag++;
		break;
	case 'b':
		bs = fstrtoll(EARGF(usage()));
		break;
	case 'c':
		cflag++;
		count = fstrtoll(EARGF(usage()));
		break;
	case 'd':
		dflag++;
		break;
	case 'f':
		strcpy(file, EARGF(usage()));
		fflag++;
		break;
	case 'n':
		nthreads = strtol(EARGF(usage()), 0, 0);
		if (nthreads < 0 || nthreads > MAXPT)
			sysfatal("invalid thread count value [0-%d].\n", MAXPT);
		break;
	case 'r':
		rflag++;
		break;
	case 's':
		ssecs = strtol(EARGF(usage()), 0, 0);
		break;
	case 't':
		total = fstrtoll(EARGF(usage()));
		break;
	case 'v':
		printf("%s\n", VER);
		return 0;
	case 'w':
		wflag++;
		break;
	case '?':
		usage();
	} ARGEND

	if (!wflag) {
		if (!rflag)
			wflag = rflag = 1;
		else if (!fflag)
			sysfatal("error: -r must be used with -f.\n");
	}

	if (fflag == 0) {
		if (argc != 1)
			usage();
		dir = *argv;
		n = strlen(dir);
		if (dir[--n] == '/')
			dir[n] = '\0';
		snprintf(file, sizeof file, "%s/ddt.%d", dir, (int) getpid());
		iofd = open(file, O_CREAT|O_RDWR|O_LARGEFILE, 0666);
		if (!Uflag)
			unlink(file);
	} else if (!wflag) {
		iofd = open(file, O_RDWR|O_LARGEFILE);
		if (iofd >= 0)
		if (!cflag) {
			fstat(iofd, &s);
			count = s.st_size / bs;
		}
	} else
		iofd = open(file, O_CREAT|O_RDWR|O_LARGEFILE, 0666);

	if (total == 0)		// -t was not used
		total = bs*count;
	else
		count = total / bs;

	memcheck(total);

	if (iofd == -1)
		sysfatal("can't open %s: %s.\n", file, strerror(errno));

	block = balloc(bs);
	if (block == nil)
		sysfatal("can't malloc %d size block for I/O.\n", bs);
	if (wflag) {
		ddw();
		if (ssecs > 0) {
			fprintf(stderr, "sleeping %d seconds ... ", ssecs);
			sleep(ssecs);
			fprintf(stderr, "done.\n");
		}
	}
	if (rflag)
		ddr();
	ddt();
	return 0;
}

