/*
	Copyright 2007 Coraid, Inc.

	Author:
	Sam Hopkins
	sah@coraid.com

	We define _FILE_OFFSET_BITS to get 64 bit pread/pwrite support.
	We define _GNU_SOURCE to obtain O_LARGEFILE, O_NOATIME.
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <sys/sysinfo.h>
#include <sys/time.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include "p9p.h"
#include "ddt.h"

enum {
	Suser,
	Snice,
	Ssystem,
	Sidle,
	Siowait,
	Sirq,
	Ssoftirq,
	Nprocstats,
	Sgtod=Nprocstats,
	Nstats,

	USER_HZ= 100,
};

char *snames[] = {
[Suser] 	= "user",
[Snice] 	= "nice",
[Ssystem] 	= "system",
[Sidle] 	= "idle",
[Siowait]	= "iowait",
[Sirq]		= "irq",
[Ssoftirq]	= "softirq",
[Sgtod]		= "gettimeofdayms"
};

#ifdef UNDEFINED

// unused, kept here for ease of future reference

// from include/asm-x86_64/msr.h and include/asm-i386/msr.h, respectively
#define rdtscll64(val) do { \
     unsigned int __a,__d; \
     asm volatile("rdtsc" : "=a" (__a), "=d" (__d)); \
     (val) = ((unsigned long)__a) | (((unsigned long)__d)<<32); \
} while(0)
#define rdtscll32(val) \
     __asm__ __volatile__("rdtsc" : "=A" (val))

unsigned long long
tsc(void)
{
	unsigned long long n = 0;

	if (sizeof(void *) == 4) {
		rdtscll32(n);
	} else {
		rdtscll64(n);
	}
	return n;
}

#endif // UNDEFINED

unsigned long long
gtod(void)
{
	struct timeval t;

typedef unsigned long long ull;

	if (gettimeofday(&t, NULL) < 0)
		return 0;
	return (ull)t.tv_sec*1000 + (ull)t.tv_usec/1000;
}

void *
stats(void *v)
{
	int i, n, fd, argc;
	char buf[128], *p, *argv[16], **args;
	unsigned long long *stats;

	if (v == nil) {
		n = sizeof *stats * Nstats;
		v = malloc(n);
		memset(v, 0, n);
	}
	stats = v;

	args = argv;
	fd = open("/proc/stat", O_RDONLY);
	if (fd < 0)
		sysfatal("procstat open failure: %s.\n", strerror(errno));
	n = read(fd, buf, sizeof buf - 1);
	if (n < 0)
		sysfatal("procstat read failure: %s.\n", strerror(errno));
	buf[n] = '\0';
	p = strchr(buf, '\n');
	if (p == nil)
		sysfatal("procstat can't find eol in '%s'.\n", buf);
	*p = '\0';
	argc = tokenize(buf, argv, nelem(argv));
	if (argc < Nprocstats + 1)
		sysfatal("procstat format assumption failure.\n");
	args++;
	for (i=0; i<Nprocstats; i++)
		stats[i] = atoi(args[i]) - stats[i];
	stats[Sgtod] = gtod() - stats[Sgtod];
	close(fd);
	return v;
}

void
rawstats(void *v)
{
	int i;
	unsigned long long *stats = v;

	for (i=0; i<Nstats; i++)
		printf("%s=%lld ", snames[i], stats[i]);
	printf("\n");
}

static int
countcpus(void)
{
	char buf[4096], *p, *r;
	int fd, n;

	fd = open("/proc/stat", O_RDONLY);
	if (fd < 0)
		sysfatal("cpus: can't open /proc/stat: %s\n", strerror(errno));
	n = read(fd, buf, sizeof buf - 1);
	if (n < 0)
		sysfatal("cpus: can't read /proc/stat: %s\n", strerror(errno));
	buf[n] = 0;

	n = -1;
	for (p=r=buf; (r=strchr(p, '\n')); p=r) {
		*r++ = '\0';
		if (strncmp(p, "cpu", 3) == 0)
			n++;
	}
	return n;
}

static char dbuf[16384];

void
cookedstats(void *v, unsigned long long bs, unsigned long long count)
{
	long long kbs, m, n;
	unsigned long cpuu;
	unsigned long long *stats = v;
	int i;
	char *p, *e;
	extern int dflag;

	for (i=0, n=0; i<Nprocstats; i++)
		n += stats[i];

	p = dbuf;
	e = p + sizeof dbuf;
	p += snprintf(p, e-p, "\tbs=%lld count=%lld stats=%lld\n", bs, count, n);

	m = stats[Sidle] + stats[Siowait];
	kbs = (bs * count);
	p += snprintf(p, e-p, "\tkbs.bsc=%lld\n", kbs);
	kbs *= 1000;
	kbs /= stats[Sgtod];
	p += snprintf(p, e-p, "\tkbs.t=%lld ncpu=%d\n", kbs, countcpus());
	kbs >>= 10;

	cpuu = (n - m) * 100 / n;

	printf("%7lld  %4ld\n", kbs, cpuu);
	if (dflag)
		printf("%*s", (int)(p-dbuf), dbuf);
}

/* check the total amount of I/O is greater than available memory */
void
memcheck(unsigned long long total)
{
	struct sysinfo s;
	unsigned long long m;

	if (sysinfo(&s) < 0) {
		perror("sysinfo failure");
		return;
	}
	m = s.totalram;
	m *= s.mem_unit;
	if (m > total)
		fprintf(stderr, "* warning: total I/O < total ram - throughput reported may not reflect I/O path.\n");
}
